@@ -51,10 +51,10 @@ struct eval_state {
 	int *ret;
 	/** Whether to quit immediately. */
 	bool *quit;
-	/** A stat() buffer, if necessary. */
-	struct bfs_stat statbuf;
 	/** A bfs_stat() buffer, if necessary. */
-	struct bfs_stat bfs_statbuf;
+	struct bfs_stat statbuf;
+	/** A bfs_stat() buffer for -xtype style tests. */
+	struct bfs_stat xstatbuf;
 };
 
 /**
@@ -76,14 +76,57 @@ static void eval_error(struct eval_state *state) {
 	}
 }
 
+#define DEBUG_FLAG(flags, flag)			\
+	if ((flags & flag) || flags == flag) {	\
+		fputs(#flag, stderr);		\
+		flags ^= flag;			\
+		if (flags) {			\
+			fputs(" | ", stderr);	\
+		}				\
+	}
+
+/**
+ * Debug stat() calls.
+ */
+static void debug_stat(const struct eval_state *state, int at_flags, enum bfs_stat_flag flags) {
+	if (!(state->cmdline->debug & DEBUG_STAT)) {
+		return;
+	}
+
+	struct BFTW *ftwbuf = state->ftwbuf;
+
+	fprintf(stderr, "bfs_stat(");
+	if (ftwbuf->at_fd == AT_FDCWD) {
+		fprintf(stderr, "AT_FDCWD");
+	} else {
+		size_t baselen = strlen(ftwbuf->path) - strlen(ftwbuf->at_path);
+		fprintf(stderr, "\"");
+		fwrite(ftwbuf->path, 1, baselen, stderr);
+		fprintf(stderr, "\"");
+	}
+
+	fprintf(stderr, ", \"%s\", ", ftwbuf->at_path);
+
+	DEBUG_FLAG(at_flags, 0);
+	DEBUG_FLAG(at_flags, AT_SYMLINK_NOFOLLOW);
+
+	fprintf(stderr, ", ");
+
+	DEBUG_FLAG(flags, 0);
+	DEBUG_FLAG(flags, BFS_STAT_BROKEN_OK);
+
+	fprintf(stderr, ")\n");
+}
+
 /**
  * Perform a bfs_stat() call if necessary.
  */
-static const struct bfs_stat *eval_bfs_stat(struct eval_state *state) {
+static const struct bfs_stat *eval_stat(struct eval_state *state) {
 	struct BFTW *ftwbuf = state->ftwbuf;
 	if (!ftwbuf->statbuf) {
-		if (bfs_stat(ftwbuf->at_fd, ftwbuf->at_path, ftwbuf->at_flags, BFS_STAT_BROKEN_OK, &state->bfs_statbuf) == 0) {
-			ftwbuf->statbuf = &state->bfs_statbuf;
+		debug_stat(state, ftwbuf->at_flags, BFS_STAT_BROKEN_OK);
+		if (bfs_stat(ftwbuf->at_fd, ftwbuf->at_path, ftwbuf->at_flags, BFS_STAT_BROKEN_OK, &state->statbuf) == 0) {
+			ftwbuf->statbuf = &state->statbuf;
 		} else {
 			eval_error(state);
 		}
@@ -91,6 +134,22 @@ static const struct bfs_stat *eval_bfs_stat(struct eval_state *state) {
 	return ftwbuf->statbuf;
 }
 
+/**
+ * Perform a bfs_stat() call for tests that flip the follow flag, like -xtype.
+ */
+static const struct bfs_stat *eval_xstat(struct eval_state *state) {
+	struct BFTW *ftwbuf = state->ftwbuf;
+	struct bfs_stat *statbuf = &state->xstatbuf;
+	if (!statbuf->mask) {
+		int at_flags = ftwbuf->at_flags ^ AT_SYMLINK_NOFOLLOW;
+		debug_stat(state, at_flags, 0);
+		if (bfs_stat(ftwbuf->at_fd, ftwbuf->at_path, at_flags, 0, statbuf) != 0) {
+			return NULL;
+		}
+	}
+	return statbuf;
+}
+
 /**
  * Get the difference (in seconds) between two struct timespecs.
  */
@@ -141,7 +200,7 @@ bool eval_access(const struct expr *expr, struct eval_state *state) {
  * Get the given timespec field out of a stat buffer.
  */
 static const struct timespec *eval_stat_time(const struct expr *expr, struct eval_state *state) {
-	const struct bfs_stat *statbuf = eval_bfs_stat(state);
+	const struct bfs_stat *statbuf = eval_stat(state);
 	if (!statbuf) {
 		return NULL;
 	}
@@ -207,7 +266,7 @@ bool eval_time(const struct expr *expr, struct eval_state *state) {
  * -used test.
  */
 bool eval_used(const struct expr *expr, struct eval_state *state) {
-	const struct bfs_stat *statbuf = eval_bfs_stat(state);
+	const struct bfs_stat *statbuf = eval_stat(state);
 	if (!statbuf) {
 		return false;
 	}
@@ -221,7 +280,7 @@ bool eval_used(const struct expr *expr, struct eval_state *state) {
  * -gid test.
  */
 bool eval_gid(const struct expr *expr, struct eval_state *state) {
-	const struct bfs_stat *statbuf = eval_bfs_stat(state);
+	const struct bfs_stat *statbuf = eval_stat(state);
 	if (!statbuf) {
 		return false;
 	}
@@ -233,7 +292,7 @@ bool eval_gid(const struct expr *expr, struct eval_state *state) {
  * -uid test.
  */
 bool eval_uid(const struct expr *expr, struct eval_state *state) {
-	const struct bfs_stat *statbuf = eval_bfs_stat(state);
+	const struct bfs_stat *statbuf = eval_stat(state);
 	if (!statbuf) {
 		return false;
 	}
@@ -245,7 +304,7 @@ bool eval_uid(const struct expr *expr, struct eval_state *state) {
  * -nogroup test.
  */
 bool eval_nogroup(const struct expr *expr, struct eval_state *state) {
-	const struct bfs_stat *statbuf = eval_bfs_stat(state);
+	const struct bfs_stat *statbuf = eval_stat(state);
 	if (!statbuf) {
 		return false;
 	}
@@ -257,7 +316,7 @@ bool eval_nogroup(const struct expr *expr, struct eval_state *state) {
  * -nouser test.
  */
 bool eval_nouser(const struct expr *expr, struct eval_state *state) {
-	const struct bfs_stat *statbuf = eval_bfs_stat(state);
+	const struct bfs_stat *statbuf = eval_stat(state);
 	if (!statbuf) {
 		return false;
 	}
@@ -277,8 +336,21 @@ bool eval_delete(const struct expr *expr, struct eval_state *state) {
 	}
 
 	int flag = 0;
-	if (ftwbuf->typeflag == BFTW_DIR) {
-		flag |= AT_REMOVEDIR;
+	if (ftwbuf->at_flags & AT_SYMLINK_NOFOLLOW) {
+		if (ftwbuf->typeflag == BFTW_DIR) {
+			flag |= AT_REMOVEDIR;
+		}
+	} else if (ftwbuf->typeflag != BFTW_LNK) {
+		// We need to know the actual type of the path, not what it points to
+		const struct bfs_stat *statbuf = eval_xstat(state);
+		if (statbuf) {
+			if (S_ISDIR(statbuf->mode)) {
+				flag |= AT_REMOVEDIR;
+			}
+		} else {
+			eval_error(state);
+			return false;
+		}
 	}
 
 	if (unlinkat(ftwbuf->at_fd, ftwbuf->at_path, flag) != 0) {
@@ -378,7 +450,7 @@ bool eval_empty(const struct expr *expr, struct eval_state *state) {
 	done_dir:
 		closedir(dir);
 	} else {
-		const struct bfs_stat *statbuf = eval_bfs_stat(state);
+		const struct bfs_stat *statbuf = eval_stat(state);
 		if (statbuf) {
 			ret = statbuf->size == 0;
 		}
@@ -392,7 +464,7 @@ done:
  * -fstype test.
  */
 bool eval_fstype(const struct expr *expr, struct eval_state *state) {
-	const struct bfs_stat *statbuf = eval_bfs_stat(state);
+	const struct bfs_stat *statbuf = eval_stat(state);
 	if (!statbuf) {
 		return false;
 	}
@@ -425,7 +497,7 @@ bool eval_nohidden(const struct expr *expr, struct eval_state *state) {
  * -inum test.
  */
 bool eval_inum(const struct expr *expr, struct eval_state *state) {
-	const struct bfs_stat *statbuf = eval_bfs_stat(state);
+	const struct bfs_stat *statbuf = eval_stat(state);
 	if (!statbuf) {
 		return false;
 	}
@@ -437,7 +509,7 @@ bool eval_inum(const struct expr *expr, struct eval_state *state) {
  * -links test.
  */
 bool eval_links(const struct expr *expr, struct eval_state *state) {
-	const struct bfs_stat *statbuf = eval_bfs_stat(state);
+	const struct bfs_stat *statbuf = eval_stat(state);
 	if (!statbuf) {
 		return false;
 	}
@@ -457,7 +529,7 @@ bool eval_lname(const struct expr *expr, struct eval_state *state) {
 		goto done;
 	}
 
-	const struct bfs_stat *statbuf = eval_bfs_stat(state);
+	const struct bfs_stat *statbuf = eval_stat(state);
 	if (!statbuf) {
 		goto done;
 	}
@@ -514,7 +586,7 @@ bool eval_path(const struct expr *expr, struct eval_state *state) {
  * -perm test.
  */
 bool eval_perm(const struct expr *expr, struct eval_state *state) {
-	const struct bfs_stat *statbuf = eval_bfs_stat(state);
+	const struct bfs_stat *statbuf = eval_stat(state);
 	if (!statbuf) {
 		return false;
 	}
@@ -548,7 +620,7 @@ bool eval_fls(const struct expr *expr, struct eval_state *state) {
 	CFILE *cfile = expr->cfile;
 	FILE *file = cfile->file;
 	const struct BFTW *ftwbuf = state->ftwbuf;
-	const struct bfs_stat *statbuf = eval_bfs_stat(state);
+	const struct bfs_stat *statbuf = eval_stat(state);
 	if (!statbuf) {
 		goto done;
 	}
@@ -640,7 +712,7 @@ error:
 bool eval_fprint(const struct expr *expr, struct eval_state *state) {
 	CFILE *cfile = expr->cfile;
 	if (cfile->colors) {
-		eval_bfs_stat(state);
+		eval_stat(state);
 	}
 
 	if (cfprintf(cfile, "%P\n", state->ftwbuf) < 0) {
@@ -667,7 +739,7 @@ bool eval_fprint0(const struct expr *expr, struct eval_state *state) {
  */
 bool eval_fprintf(const struct expr *expr, struct eval_state *state) {
 	if (expr->printf->needs_stat) {
-		if (!eval_bfs_stat(state)) {
+		if (!eval_stat(state)) {
 			goto done;
 		}
 	}
@@ -772,7 +844,7 @@ bool eval_regex(const struct expr *expr, struct eval_state *state) {
  * -samefile test.
  */
 bool eval_samefile(const struct expr *expr, struct eval_state *state) {
-	const struct bfs_stat *statbuf = eval_bfs_stat(state);
+	const struct bfs_stat *statbuf = eval_stat(state);
 	if (!statbuf) {
 		return false;
 	}
@@ -784,7 +856,7 @@ bool eval_samefile(const struct expr *expr, struct eval_state *state) {
  * -size test.
  */
 bool eval_size(const struct expr *expr, struct eval_state *state) {
-	const struct bfs_stat *statbuf = eval_bfs_stat(state);
+	const struct bfs_stat *statbuf = eval_stat(state);
 	if (!statbuf) {
 		return false;
 	}
@@ -809,7 +881,7 @@ bool eval_size(const struct expr *expr, struct eval_state *state) {
  * -sparse test.
  */
 bool eval_sparse(const struct expr *expr, struct eval_state *state) {
-	const struct bfs_stat *statbuf = eval_bfs_stat(state);
+	const struct bfs_stat *statbuf = eval_stat(state);
 	if (!statbuf) {
 		return false;
 	}
@@ -837,21 +909,16 @@ bool eval_xtype(const struct expr *expr, struct eval_state *state) {
 		return eval_type(expr, state);
 	}
 
-	// -xtype does the opposite of everything else
-	int at_flags = ftwbuf->at_flags ^ AT_SYMLINK_NOFOLLOW;
-
-	struct bfs_stat sb;
-	if (bfs_stat(ftwbuf->at_fd, ftwbuf->at_path, at_flags, 0, &sb) != 0) {
-		if (!follow && is_nonexistence_error(errno)) {
-			// Broken symlink
-			return eval_type(expr, state);
-		} else {
-			eval_error(state);
-			return false;
-		}
+	const struct bfs_stat *statbuf = eval_xstat(state);
+	if (statbuf) {
+		return mode_to_typeflag(statbuf->mode) & expr->idata;
+	} else if (!follow && is_nonexistence_error(errno)) {
+		// Broken symlink
+		return eval_type(expr, state);
+	} else {
+		eval_error(state);
+		return false;
 	}
-
-	return mode_to_typeflag(sb.mode) & expr->idata;
 }
 
 #if _POSIX_MONOTONIC_CLOCK > 0
@@ -977,33 +1044,6 @@ bool eval_comma(const struct expr *expr, struct eval_state *state) {
 	return eval_expr(expr->rhs, state);
 }
 
-/**
- * Debug stat() calls.
- */
-static void debug_stat(const struct eval_state *state) {
-	struct BFTW *ftwbuf = state->ftwbuf;
-
-	fprintf(stderr, "fstatat(");
-	if (ftwbuf->at_fd == AT_FDCWD) {
-		fprintf(stderr, "AT_FDCWD");
-	} else {
-		size_t baselen = strlen(ftwbuf->path) - strlen(ftwbuf->at_path);
-		fprintf(stderr, "\"");
-		fwrite(ftwbuf->path, 1, baselen, stderr);
-		fprintf(stderr, "\"");
-	}
-
-	fprintf(stderr, ", \"%s\", ", ftwbuf->at_path);
-
-	if (ftwbuf->at_flags == AT_SYMLINK_NOFOLLOW) {
-		fprintf(stderr, "AT_SYMLINK_NOFOLLOW");
-	} else {
-		fprintf(stderr, "%d", ftwbuf->at_flags);
-	}
-
-	fprintf(stderr, ")\n");
-}
-
 /**
  * Dump the bftw_typeflag for -D search.
  */
@@ -1083,11 +1123,16 @@ static enum bftw_action cmdline_callback(struct BFTW *ftwbuf, void *ptr) {
 	state.action = BFTW_CONTINUE;
 	state.ret = &args->ret;
 	state.quit = &args->quit;
+	state.xstatbuf.mask = 0;
+
+	if (ftwbuf->statbuf) {
+		debug_stat(&state, ftwbuf->at_flags, BFS_STAT_BROKEN_OK);
+	}
 
 	if (ftwbuf->typeflag == BFTW_ERROR) {
 		if (!eval_should_ignore(&state, ftwbuf->error)) {
 			args->ret = EXIT_FAILURE;
-			cfprintf(cmdline->cerr, "%{er}error: '%s': %m%{rs}\n", ftwbuf->path);
+			cfprintf(cmdline->cerr, "%{er}error: '%s': %s%{rs}\n", ftwbuf->path, strerror(ftwbuf->error));
 		}
 		state.action = BFTW_SKIP_SUBTREE;
 		goto done;
@@ -1119,10 +1164,6 @@ static enum bftw_action cmdline_callback(struct BFTW *ftwbuf, void *ptr) {
 	}
 
 done:
-	if ((cmdline->debug & DEBUG_STAT) && ftwbuf->statbuf) {
-		debug_stat(&state);
-	}
-
 	if (cmdline->debug & DEBUG_SEARCH) {
 		fprintf(stderr,
 		        "cmdline_callback({ "
@@ -1198,22 +1239,14 @@ static int infer_fdlimit(const struct cmdline *cmdline) {
  * Dump the bftw() flags for -D search.
  */
 static void dump_bftw_flags(enum bftw_flags flags) {
-#define DUMP_BFTW_FLAG(flag)			\
-	if (flags & flag) {			\
-		fputs(#flag, stderr);		\
-		flags ^= flag;			\
-		if (flags) {			\
-			fputs(" | ", stderr);	\
-		}				\
-	}
-
-	DUMP_BFTW_FLAG(BFTW_STAT);
-	DUMP_BFTW_FLAG(BFTW_RECOVER);
-	DUMP_BFTW_FLAG(BFTW_DEPTH);
-	DUMP_BFTW_FLAG(BFTW_COMFOLLOW);
-	DUMP_BFTW_FLAG(BFTW_LOGICAL);
-	DUMP_BFTW_FLAG(BFTW_DETECT_CYCLES);
-	DUMP_BFTW_FLAG(BFTW_XDEV);
+	DEBUG_FLAG(flags, 0);
+	DEBUG_FLAG(flags, BFTW_STAT);
+	DEBUG_FLAG(flags, BFTW_RECOVER);
+	DEBUG_FLAG(flags, BFTW_DEPTH);
+	DEBUG_FLAG(flags, BFTW_COMFOLLOW);
+	DEBUG_FLAG(flags, BFTW_LOGICAL);
+	DEBUG_FLAG(flags, BFTW_DETECT_CYCLES);
+	DEBUG_FLAG(flags, BFTW_XDEV);
 
 	assert(!flags);
 }
