@@ -90,13 +90,16 @@ static size_t bfs_exec_arg_max(const struct bfs_exec *execbuf) {
 	bfs_exec_debug(execbuf, "ARG_MAX: %ld remaining after fixed arguments\n", arg_max);
 
 	// Assume arguments are counted with the granularity of a single page,
-	// and allow two pages of headroom to account for rounding as well as
-	// any other data we may not be counting
+	// so allow a one page cushion to account for rounding up
 	long page_size = sysconf(_SC_PAGESIZE);
 	if (page_size < 4096) {
 		page_size = 4096;
 	}
-	arg_max -= 2*page_size;
+	arg_max -= page_size;
+	bfs_exec_debug(execbuf, "ARG_MAX: %ld remaining after page cushion\n", arg_max);
+
+	// POSIX recommends an additional 2048 bytes of headroom
+	arg_max -= 2048;
 	bfs_exec_debug(execbuf, "ARG_MAX: %ld remaining after headroom\n", arg_max);
 
 	if (arg_max < 0) {
@@ -360,17 +363,28 @@ static int bfs_exec_spawn(const struct bfs_exec *execbuf) {
 			return -1;
 		}
 
-		int status;
-		if (waitpid(pid, &status, 0) < 0) {
+		int wstatus;
+		if (waitpid(pid, &wstatus, 0) < 0) {
 			return -1;
 		}
 
 		errno = 0;
-		if (WIFEXITED(status) && WEXITSTATUS(status) == EXIT_SUCCESS) {
-			return 0;
+
+		if (WIFEXITED(wstatus)) {
+			int status = WEXITSTATUS(wstatus);
+			if (status == EXIT_SUCCESS) {
+				return 0;
+			} else {
+				bfs_exec_debug(execbuf, "Command '%s' failed with status %d\n", execbuf->argv[0], status);
+			}
+		} else if (WIFSIGNALED(wstatus)) {
+			int sig = WTERMSIG(wstatus);
+			bfs_exec_debug(execbuf, "Command '%s' terminated by signal %d\n", execbuf->argv[0], sig);
 		} else {
-			return -1;
+			bfs_exec_debug(execbuf, "Command '%s' terminated abnormally\n", execbuf->argv[0]);
 		}
+
+		return -1;
 	} else {
 		// Child
 		close(pipefd[0]);
@@ -446,13 +460,14 @@ static bool bfs_exec_args_remain(const struct bfs_exec *execbuf) {
 
 /** Execute the pending command from a BFS_EXEC_MULTI execbuf. */
 static int bfs_exec_flush(struct bfs_exec *execbuf) {
-	int ret = 0;
+	int ret = 0, error = 0;
 
 	size_t orig_argc = execbuf->argc;
 	while (bfs_exec_args_remain(execbuf)) {
 		execbuf->argv[execbuf->argc] = NULL;
 		ret = bfs_exec_spawn(execbuf);
-		if (ret == 0 || errno != E2BIG) {
+		error = errno;
+		if (ret == 0 || error != E2BIG) {
 			break;
 		}
 
@@ -464,30 +479,30 @@ static int bfs_exec_flush(struct bfs_exec *execbuf) {
 		--execbuf->argc;
 	}
 	size_t new_argc = execbuf->argc;
-
-	int error = errno;
+	size_t new_size = execbuf->arg_size;
 
 	for (size_t i = execbuf->tmpl_argc - 1; i < new_argc; ++i) {
 		free(execbuf->argv[i]);
 	}
 	execbuf->argc = execbuf->tmpl_argc - 1;
+	execbuf->arg_size = 0;
 
 	if (new_argc < orig_argc) {
+		execbuf->arg_max = new_size;
+		bfs_exec_debug(execbuf, "ARG_MAX: %zu\n", execbuf->arg_max);
+
 		// If we recovered from E2BIG, there are unused arguments at the
 		// end of the list
 		for (size_t i = new_argc + 1; i <= orig_argc; ++i) {
 			if (error == 0) {
 				execbuf->argv[execbuf->argc] = execbuf->argv[i];
+				execbuf->arg_size += bfs_exec_arg_size(execbuf->argv[execbuf->argc]);
 				++execbuf->argc;
 			} else {
 				free(execbuf->argv[i]);
 			}
 		}
-
-		execbuf->arg_max = execbuf->arg_size;
-		bfs_exec_debug(execbuf, "ARG_MAX: %zu\n", execbuf->arg_max);
 	}
-	execbuf->arg_size = 0;
 
 	errno = error;
 	return ret;
@@ -597,6 +612,9 @@ int bfs_exec_finish(struct bfs_exec *execbuf) {
 		while (bfs_exec_args_remain(execbuf)) {
 			execbuf->ret |= bfs_exec_flush(execbuf);
 		}
+		if (execbuf->ret != 0) {
+			bfs_exec_debug(execbuf, "One or more executions of '%s' failed\n", execbuf->argv[0]);
+		}
 	}
 	return execbuf->ret;
 }
