@@ -24,10 +24,22 @@ export TZ=UTC
 
 # The temporary directory that will hold our test data
 TMP="$(mktemp -d "${TMPDIR:-/tmp}"/bfs.XXXXXXXXXX)"
-chown "$(id -u)":"$(id -g)" "$TMP"
+chown "$(id -u):$(id -g)" "$TMP"
 
 # Clean up temporary directories on exit
+CLEAN=yes
 function cleanup() {
+    if [ ! "$CLEAN" ]; then
+        return
+    fi
+
+    # Don't force rm to deal with long paths
+    for dir in "$TMP"/deep/*/*; do
+        if [ -d "$dir" ]; then
+            (cd "$dir" && rm -rf *)
+        fi
+    done
+
     rm -rf "$TMP"
 }
 trap cleanup EXIT
@@ -164,6 +176,7 @@ posix_tests=(
     test_depth_mindepth_2
     test_depth_maxdepth_1
     test_depth_maxdepth_2
+    test_depth_error
     test_name
     test_name_root
     test_name_root_depth
@@ -177,8 +190,11 @@ posix_tests=(
     test_H
     test_H_slash
     test_H_broken
+    test_H_notdir
     test_H_newer
     test_L
+    test_L_broken
+    test_L_notdir
     test_L_depth
     test_user_name
     test_user_id
@@ -223,8 +239,13 @@ bsd_tests=(
     test_samefile
     test_samefile_symlink
     test_H_samefile_symlink
+    test_L_samefile_symlink
     test_samefile_broken
     test_H_samefile_broken
+    test_L_samefile_broken
+    test_samefile_notdir
+    test_H_samefile_notdir
+    test_L_samefile_notdir
     test_name_slash
     test_name_slashes
     test_iname
@@ -245,6 +266,7 @@ bsd_tests=(
     test_ok_stdin
     test_okdir_stdin
     test_delete
+    test_L_delete
     test_rm
     test_regex
     test_iregex
@@ -301,8 +323,13 @@ gnu_tests=(
     test_samefile
     test_samefile_symlink
     test_H_samefile_symlink
+    test_L_samefile_symlink
     test_samefile_broken
     test_H_samefile_broken
+    test_L_samefile_broken
+    test_samefile_notdir
+    test_H_samefile_notdir
+    test_L_samefile_notdir
     test_xtype_l
     test_xtype_f
     test_L_xtype_l
@@ -343,6 +370,7 @@ gnu_tests=(
     test_perm_symbolic_slash
     test_perm_leading_plus_symbolic_slash
     test_delete
+    test_L_delete
     test_regex
     test_iregex
     test_regex_parens
@@ -440,6 +468,9 @@ for arg; do
             GNU=yes
             ALL=yes
             ;;
+        --noclean)
+            CLEAN=
+            ;;
         --update)
             UPDATE=yes
             ;;
@@ -457,7 +488,11 @@ for arg; do
     esac
 done
 
-if [ -z "$EXPLICIT" ]; then
+if [ ! "$CLEAN" ]; then
+    echo "Test files saved to $TMP"
+fi
+
+if [ ! "$EXPLICIT" ]; then
     enable_tests "${posix_tests[@]}"
     [ "$BSD" ] && enable_tests "${bsd_tests[@]}"
     [ "$GNU" ] && enable_tests "${gnu_tests[@]}"
@@ -516,11 +551,17 @@ function bfs_diff() (
     # substitution, even with low ulimit -n
     exec 3>&-
 
-    local OUT="$TESTS/${FUNCNAME[1]}.out"
+    local EXPECTED="$TESTS/${FUNCNAME[1]}.out"
     if [ "$UPDATE" ]; then
-        $BFS "$@" | bfs_sort >"$OUT"
+        local ACTUAL="$EXPECTED"
     else
-        diff -u "$OUT" <($BFS "$@" | bfs_sort)
+        local ACTUAL="$TMP/${FUNCNAME[1]}.out"
+    fi
+
+    $BFS "$@" | bfs_sort >"$ACTUAL"
+
+    if [ ! "$UPDATE" ]; then
+        diff -u "$EXPECTED" "$ACTUAL"
     fi
 )
 
@@ -595,6 +636,20 @@ function test_depth_maxdepth_2() {
     bfs_diff basic -maxdepth 2 -depth
 }
 
+function test_depth_error() {
+    rm -rf scratch/*
+    touchp scratch/foo/bar
+    chmod -r scratch/foo
+
+    bfs_diff scratch -depth 2>/dev/null
+    local ret=$?
+
+    chmod +r scratch/foo
+    rm -rf scratch/*
+
+    return $ret
+}
+
 function test_name() {
     bfs_diff basic -name '*f*'
 }
@@ -715,6 +770,10 @@ function test_H_broken() {
     bfs_diff -H links/broken
 }
 
+function test_H_notdir() {
+    bfs_diff -H links/notdir
+}
+
 function test_H_newer() {
     bfs_diff -H times -newer times/l
 }
@@ -723,6 +782,14 @@ function test_L() {
     bfs_diff -L links 2>/dev/null
 }
 
+function test_L_broken() {
+    bfs_diff -H links/broken
+}
+
+function test_L_notdir() {
+    bfs_diff -H links/notdir
+}
+
 function test_X() {
     bfs_diff -X weirdnames 2>/dev/null
 }
@@ -747,6 +814,10 @@ function test_H_samefile_symlink() {
     bfs_diff -H links -samefile links/symlink
 }
 
+function test_L_samefile_symlink() {
+    bfs_diff -L links -samefile links/symlink 2>/dev/null
+}
+
 function test_samefile_broken() {
     bfs_diff links -samefile links/broken
 }
@@ -755,6 +826,22 @@ function test_H_samefile_broken() {
     bfs_diff -H links -samefile links/broken
 }
 
+function test_L_samefile_broken() {
+    bfs_diff -L links -samefile links/broken 2>/dev/null
+}
+
+function test_samefile_notdir() {
+    bfs_diff links -samefile links/notdir
+}
+
+function test_H_samefile_notdir() {
+    bfs_diff -H links -samefile links/notdir
+}
+
+function test_L_samefile_notdir() {
+    bfs_diff -L links -samefile links/notdir 2>/dev/null
+}
+
 function test_xtype_l() {
     bfs_diff links -xtype l 2>/dev/null
 }
@@ -1092,6 +1179,18 @@ function test_delete() {
     bfs_diff scratch
 }
 
+function test_L_delete() {
+    rm -rf scratch/*
+    mkdir scratch/foo
+    mkdir scratch/bar
+    ln -s ../foo scratch/bar/baz
+
+    # Don't try to rmdir() a symlink
+    invoke_bfs -L scratch/bar -delete || return 1
+
+    bfs_diff scratch
+}
+
 function test_rm() {
     rm -rf scratch/*
     touchp scratch/foo/bar/baz
@@ -1227,7 +1326,7 @@ function test_quit_implicit_print() {
 }
 
 function test_inum() {
-    local inode="$(ls -id basic/k/foo/bar | cut -f1 -d' ')"
+    local inode="$(ls -id basic/k/foo/bar | awk '{ print $1 }')"
     bfs_diff basic -inum "$inode"
 }
 
@@ -1282,12 +1381,17 @@ function test_printf_leak() {
 
 function test_printf_nul() {
     # NUL byte regression test
-    local OUT="$TESTS/${FUNCNAME[0]}.out"
-    local ARGS=(basic -maxdepth 0 -printf '%h\0%f\n')
+    local EXPECTED="$TESTS/${FUNCNAME[0]}.out"
     if [ "$UPDATE" ]; then
-        invoke_bfs "${ARGS[@]}" >"$OUT"
+        local ACTUAL="$EXPECTED"
     else
-        diff -u "$OUT" <(invoke_bfs "${ARGS[@]}")
+        local ACTUAL="$TMP/${FUNCNAME[0]}.out"
+    fi
+
+    invoke_bfs basic -maxdepth 0 -printf '%h\0%f\n' >"$ACTUAL"
+
+    if [ ! "$UPDATE" ]; then
+        diff -u "$EXPECTED" "$ACTUAL"
     fi
 }
 
@@ -1297,7 +1401,7 @@ function test_printf_w() {
 }
 
 function test_fstype() {
-    fstype="$(invoke_bfs -printf '%F\n' | head -n1)"
+    fstype="$(invoke_bfs basic -maxdepth 0 -printf '%F\n')"
     bfs_diff basic -fstype "$fstype"
 }
 
